@@ -1,4 +1,15 @@
 #define EIGEN_DONT_PARALLELIZE // see https://github.com/kaskr/adcomp/issues/390
+#define DISABLE_AD if(isDouble<Type>::value)
+#define NAN std::numeric_limits<double>::quiet_NaN()
+// set correlation matrix for output, unless corr output has been disabled
+// (make this a function instead?
+#define SET_COR    if (term.fullCor == 1) { \
+      term.corr = nldens.cov(); \
+    } else { \
+      term.corr.resize(1,1); \
+      term.corr(0,0) = NAN; \
+    }
+
 #include <TMB.hpp>
 #include <R_ext/Error.h>
 #include "init.h"
@@ -77,7 +88,12 @@ enum valid_covStruct {
   toep_covstruct = 8,
   rr_covstruct = 9,
   homdiag_covstruct = 10,
-  propto_covstruct = 11
+  propto_covstruct = 11,
+  // should perhaps be next to homdiag but don't want to mess
+  //  up interpretation of stored fits ...
+  hetar1_covstruct = 12,
+  homcs_covstruct = 13,
+  homtoep_covstruct = 14
 };
 
 // should probably be named just 'predictCode';
@@ -161,6 +177,38 @@ Type inverse_linkfun(Type eta, int link) {
   return ans;
 }
 
+template<class Type>
+Type linkfun(Type mu, int link) {
+  Type ans;
+  switch (link) {
+  case log_link:
+    ans = log(mu);
+    break;
+  case identity_link:
+    ans = mu;
+    break;
+  case logit_link:
+    ans = logit(mu);
+    break;
+  case probit_link:
+    ans = qnorm(mu);
+    break;
+  case cloglog_link:
+    ans = log(-log(Type(1)-mu));
+    break;
+  case inverse_link:
+    ans = Type(1) / mu;
+    break;
+  case sqrt_link:
+    ans = sqrt(mu);
+    break;
+    // TODO: Implement remaining links
+  default:
+    error("Link not implemented!");
+  } // End switch
+  return ans;
+}
+
 /* logit transformed inverse_linkfun without losing too much
    accuracy */
 template<class Type>
@@ -257,6 +305,7 @@ struct per_term_info {
   int blockReps;     // Repeat block number of times
   int blockNumTheta; // Parameter count per block
   int simCode;       // Simulation code (zero, fixed, or draw new random deviate?)
+  int fullCor;       // Compute/store full correlation matrix?
   matrix<Type> dist;
   vector<Type> times;// For ar1 case
   // Report output
@@ -276,11 +325,13 @@ struct terms_t : vector<per_term_info<Type> > {
       int blockReps = (int) REAL(getListElement(y, "blockReps", &isNumericScalar))[0];
       int blockNumTheta = (int) REAL(getListElement(y, "blockNumTheta", &isNumericScalar))[0];
       int simCode = (int) REAL(getListElement(y, "simCode", &isNumericScalar))[0];
+      int fullCor = (int) REAL(getListElement(y, "fullCor", &isNumericScalar))[0];
       (*this)(i).blockCode = blockCode;
       (*this)(i).blockSize = blockSize;
       (*this)(i).blockReps = blockReps;
       (*this)(i).blockNumTheta = blockNumTheta;
       (*this)(i).simCode = simCode;
+      (*this)(i).fullCor = fullCor;
       // Optionally, pass time vector:
       SEXP t = getListElement(y, "times");
       if(!Rf_isNull(t)){
@@ -380,112 +431,164 @@ Type termwise_nll(array<Type> &U, vector<Type> theta, per_term_info<Type>& term,
         
       }
     }
-    term.corr = nldens.cov(); // For report
+
+    // FIXME: DRY/make this into a function or macro
+    SET_COR;
+    
     term.sd = sd;             // For report
   }
-  else if (term.blockCode == cs_covstruct){
-    // case: cs_covstruct
+  else if (term.blockCode == cs_covstruct || term.blockCode == homcs_covstruct) {
+    // case: cs_covstruct or homcs_covstruct
     int n = term.blockSize;
-    vector<Type> logsd = theta.head(n);
-    Type corr_transf = theta(n);
+    vector<Type> logsd(n);
+    for (int i = 0; i < n; i++) {
+      if (term.blockCode == cs_covstruct) {
+        logsd(i) = theta(i);
+      } else {
+        logsd(i) = theta(0);
+      }
+    }
+    Type corr_transf = theta.tail(1)(0);
     vector<Type> sd = exp(logsd);
     Type a = Type(1) / (Type(n) - Type(1));
     Type rho = invlogit(corr_transf) * (Type(1) + a) - a;
     matrix<Type> corr(n,n);
     for(int i=0; i<n; i++)
       for(int j=0; j<n; j++)
-	corr(i,j) = (i==j ? Type(1) : rho);
+        corr(i,j) = (i==j ? Type(1) : rho);
     density::MVNORM_t<Type> nldens(corr);
     density::VECSCALE_t<density::MVNORM_t<Type> > scnldens = density::VECSCALE(nldens, sd);
     for(int i = 0; i < term.blockReps; i++){
       ans += scnldens(U.col(i));
       if (do_simulate) {
-	if (term.simCode != random_simcode) {
-	  Rf_error("simcode not yet implemented for cs cov struct");
-	}
+        if (term.simCode != random_simcode) {
+          Rf_error("simcode not yet implemented for cs cov struct");
+        }
         U.col(i) = sd * nldens.simulate();
       }
     }
-    term.corr = nldens.cov(); // For report
+    SET_COR;
     term.sd = sd;             // For report
   }
-  else if (term.blockCode == toep_covstruct){
-    // case: toep_covstruct
+  else if (term.blockCode == toep_covstruct || term.blockCode == homtoep_covstruct) {
+    // case: toep_covstruct or homtoep_covstruct
     int n = term.blockSize;
-    vector<Type> logsd = theta.head(n);
+    vector<Type> logsd(n);
+    for (int i = 0; i < n; i++) {
+      if (term.blockCode == toep_covstruct) {
+        logsd(i) = theta(i);
+      } else {
+        logsd(i) = theta(0);
+      }
+    }
     vector<Type> sd = exp(logsd);
-    vector<Type> parms = theta.tail(n-1);              // Corr parms
-    parms = parms / sqrt(Type(1.0) + parms * parms );  // Now in (-1,1)
+    vector<Type> corr_params = theta.tail(n-1);
+    corr_params = corr_params / sqrt(Type(1.0) + corr_params * corr_params);
     matrix<Type> corr(n,n);
-    for(int i=0; i<n; i++)
+    for (int i=0; i<n; i++)
       for(int j=0; j<n; j++)
         corr(i,j) = (i==j ? Type(1) :
-                     parms( (i > j ? i-j : j-i) - 1 ) );
+                     corr_params( (i > j ? i-j : j-i) - 1 ) );
     density::MVNORM_t<Type> nldens(corr);
     density::VECSCALE_t<density::MVNORM_t<Type> > scnldens = density::VECSCALE(nldens, sd);
     for(int i = 0; i < term.blockReps; i++){
       ans += scnldens(U.col(i));
       if (do_simulate) {
-	if (term.simCode != random_simcode) {
-	  Rf_error("simcode not yet implemented for toep cov struct");
-	}
+        if (term.simCode != random_simcode) {
+          Rf_error("simcode not yet implemented for toep and homtoep cov struct");
+        }
         U.col(i) = sd * nldens.simulate();
       }
     }
-    term.corr = nldens.cov(); // For report
+    SET_COR;
     term.sd = sd;             // For report
   }
-  else if (term.blockCode == ar1_covstruct){
+  else if (term.blockCode == ar1_covstruct ||
+	   term.blockCode == hetar1_covstruct) {
     // case: ar1_covstruct
     //  * NOTE: Valid parameter space is phi in [-1, 1]
     //  * NOTE: 'times' not used as we assume unit distance between consecutive time points.
     int n = term.blockSize;
-    Type logsd = theta(0);
-    Type corr_transf = theta(1);
+    vector<Type> logsd = theta.head(term.blockNumTheta-1);
+    Type corr_transf = theta(term.blockNumTheta-1);
     Type phi = corr_transf / sqrt(1.0 + pow(corr_transf, 2));
-    Type sd = exp(logsd);
+    vector<Type> sd = exp(logsd);
     
     for(int j = 0; j < term.blockReps; j++){
-       for(int i=1; i<n; i++){
-          //Type phi = corr_transf / sqrt(1.0 + pow(corr_transf, 2));
-          ans -= dnorm(U(i, j), phi * U(i-1, j), sd * sqrt(1 - phi*phi), true);
+      // Initialize
+      if (term.blockCode == hetar1_covstruct) {
+        ans -= dnorm(U(0, j) / sd(0), Type(0), Type(1), true);
+        ans += logsd(0);
+      } else { // ar1_covstruct
+        ans -= dnorm(U(0, j), Type(0), sd(0), true);
       }
-      ans -= dnorm(U(0, j), Type(0), sd, true);   // Initialize
       if (do_simulate) {
         switch(term.simCode) {
           case fix_simcode:
-          // do nothing, leave U values as is
-             break;
+            break;
           case zero_simcode:
-            for (int i=0; i < U.rows(); i++) {
-              U(i,j) = Type(0);
-              };
+            U(0,j) = Type(0);
             break;
           case random_simcode:
-            U(0, j) = rnorm(Type(0), sd);
-            for(int i=1; i<n; i++){
-              U(i, j) = rnorm( phi * U(i-1, j), sd * sqrt(1 - phi*phi) );
-              }
+            U(0, j) = rnorm(Type(0), sd(0));
             break;
-           default: error ("unknown simcode");
-           }
-      } // do_simulate
-  }
-  
-  
-    // For consistency with output for other structs we report entire
-    // covariance matrix.
-    if(isDouble<Type>::value) { // Disable AD for this part
-      term.corr.resize(n,n);
-      term.sd.resize(n);
-      for(int i=0; i<n; i++){
-	term.sd(i) = sd;
-	for(int j=0; j<n; j++){
-	  term.corr(i,j) = pow(phi, abs(i-j));
-	}
+        }
       }
-    }
-  }
+      for(int i=1; i<n; i++) {
+
+        if (term.blockCode == hetar1_covstruct) {
+          ans -= dnorm(U(i, j) / sd(i), phi * U(i-1, j) / sd(i-1), sqrt(1 - phi*phi), true);
+          ans += logsd(i);
+        } else { // ar1_covstruct
+          ans -= dnorm(U(i, j), phi * U(i-1, j), sd(0) * sqrt(1 - phi*phi), true);
+        }
+      
+        if (do_simulate) {
+	        switch(term.simCode) {
+            case fix_simcode:
+              // do nothing, leave U values as is
+	            break;
+            case zero_simcode:
+              for (int i=0; i < U.rows(); i++) {
+                U(i, j) = Type(0);
+	            };
+              break;
+            case random_simcode:
+              if (term.blockCode == hetar1_covstruct) {
+                  U(i, j) = sd(i) * rnorm(phi * U(i-1, j) / sd(i-1), sqrt(1 - phi*phi));
+              } else {
+                  U(i, j) = rnorm(phi * U(i-1, j), sd(0) * sqrt(1 - phi*phi));
+              }
+              break;
+            default: error ("unknown simcode");
+	        } // term.simCode
+      	} // do_simulate
+      } // loop over lags
+    } // loop over blocks
+    DISABLE_AD { // Disable AD for this part
+      if (term.fullCor == 0) {
+        // report *only* phi in the corr struct
+        term.corr.resize(1,1);
+        term.corr(0,0) = phi;
+      } else {
+        term.corr.resize(n,n);
+        for (int i=0; i<n; i++) {
+          for(int j=0; j<n; j++){
+            term.corr(i,j) = pow(phi, abs(i-j));
+          }
+        }
+      }
+      if (term.blockCode == hetar1_covstruct) {
+        term.sd.resize(n);
+        for(int i=0; i<n; i++){
+          term.sd(i) = sd(i);
+        }
+      } else { // ar1_covstruct
+        term.sd.resize(1);
+        term.sd(0) = sd(0);
+      }
+    } // DISABLE_AD 
+  } // [het]ar1_covstruct
   else if (term.blockCode == ou_covstruct){
     // case: ou_covstruct
     //  * NOTE: this is the continuous time version of ar1.
@@ -521,22 +624,30 @@ Type termwise_nll(array<Type> &U, vector<Type> theta, per_term_info<Type>& term,
         default: error ("unknown simcode");
         }
       } // do_simulate
+
       
     }
-    // For consistency with output for other structs we report entire
-    // covariance matrix.
-    if(isDouble<Type>::value) { // Disable AD for this part
-      term.corr.resize(n,n);
-      term.sd.resize(n);
-      for(int i=0; i<n; i++){
-	term.sd(i) = sd;
-	for(int j=0; j<n; j++){
-	  term.corr(i,j) =
-	    exp(-exp(corr_transf) * CppAD::abs(term.times(i) - term.times(j)));
-	}
+    // Report only the sd and the decay parameter
+    DISABLE_AD { // Disable AD for this part
+      if (term.fullCor==1) {
+        term.corr.resize(n,n);
+        term.sd.resize(n);
+        for(int i=0; i<n; i++) {
+          term.sd(i) = sd;
+          for(int j=0; j<n; j++){
+            term.corr(i,j) =
+              exp(-exp(corr_transf) * CppAD::abs(term.times(i) - term.times(j)));
+          }
+        }
+      } else {
+        term.corr.resize(1,1);
+        term.sd.resize(1);
+        term.sd(0) = sd;
+        term.corr(0,0) = exp(corr_transf);
       }
     }
-  }
+  } // OU covstruct
+  
   // Spatial correlation structures
   else if (term.blockCode == exp_covstruct ||
            term.blockCode == gau_covstruct ||
@@ -579,7 +690,7 @@ Type termwise_nll(array<Type> &U, vector<Type> theta, per_term_info<Type>& term,
         U.col(i) = sd * nldens.simulate();
       }
     }
-    term.corr = corr;   // For report
+    if (term.fullCor==1) term.corr = corr;   // For report
     term.sd.resize(n);  // For report
     term.sd.fill(sd);
   }
@@ -633,18 +744,20 @@ Type termwise_nll(array<Type> &U, vector<Type> theta, per_term_info<Type>& term,
     // (i.e. user is assumed to be passing the 'non-spherical' latent variables)
     if (term.simCode != fix_simcode) {
       for(int i = 0; i < term.blockReps; i++){
-	vector<Type> usub = U.col(i).segment(0, rank);
-	U.col(i) = Lambda * usub;
+        vector<Type> usub = U.col(i).segment(0, rank);
+        U.col(i) = Lambda * usub;
       }
     }
 
     // computing the correlation matrix and std devs
     // (the same D^(-1/2) L L^T D^(-1/2) transformation that we use for correlations
     term.fact_load = Lambda;
-    if(isDouble<Type>::value) {
-      term.corr = Lambda * Lambda.transpose();
-      term.sd = term.corr.diagonal().array().sqrt();
-      term.corr.array() /= (term.sd.matrix() * term.sd.matrix().transpose()).array();
+    if (term.fullCor==1) {
+      DISABLE_AD {
+        term.corr = Lambda * Lambda.transpose();
+        term.sd = term.corr.diagonal().array().sqrt();
+        term.corr.array() /= (term.sd.matrix() * term.sd.matrix().transpose()).array();
+      }
     }
   }
   else if (term.blockCode == propto_covstruct){
@@ -662,7 +775,7 @@ Type termwise_nll(array<Type> &U, vector<Type> theta, per_term_info<Type>& term,
         U.col(i) = sd * nldens.simulate();
       }
     }
-    term.corr = nldens.cov(); // For report
+    SET_COR;
     term.sd = sd;             // For report
   }
   else error("covStruct not implemented!");
@@ -1222,6 +1335,22 @@ Type objective_function<Type>::operator() ()
   vector<Type> mu_predict = mu(whichPredict);
   vector<Type> eta_predict = eta(whichPredict);
 
+  DATA_FACTOR(aggregate);
+  if (aggregate.size() > 0) {
+    if (aggregate.size() != mu_predict.size())
+      Rf_error("'aggregate' wrong size");
+    vector<Type> tmp(NLEVELS(aggregate));
+    tmp.setZero();
+    for (int i=0; i<aggregate.size(); i++) {
+      tmp[aggregate[i]] += mu_predict[i];
+    }
+    mu_predict = tmp;
+    for (int i=0; i<tmp.size(); i++) {
+      tmp[i] = linkfun(tmp[i], link);
+    }
+    eta_predict = tmp;
+  }
+
   REPORT(mu_predict);
   REPORT(eta_predict);
   // ADREPORT expensive for long vectors - only needed by predict() method
