@@ -21,3 +21,16 @@ test_that("diagnose works with Tweedie fits", {
     cc <- capture.output(d <- diagnose(fm1, explain = FALSE))
     expect_false(d)
 })
+
+test_that("diagnose with mapped parameters", {
+    ## GH 1120
+    M0 <- suppressWarnings(
+        glmmTMB(
+            Reaction ~ Days + cs(0 + factor(Days) | Subject), dispformula = ~ 0,
+            map = list(theta = factor(c(rep(1, 10), 2))),
+            data = sleepstudy, REML = TRUE)
+    )
+    dd_out <- capture.output(dd <- diagnose(M0))
+    expect_equal(dd, FALSE)
+    expect_true(any(grepl("Unusually large Z-statistics", dd_out)))
+})
